/*
 * Copyright (c) 2001 - 2003 Bernd Walter Computer Technology
 * http://www.bwct.de
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $URL: https://devel.bwct.de/svn/projects/modbus/mb_magcardd/tags/0.1/mb_magcardd.c $
 * $Date: 2004-11-09 19:43:11 +0100 (Tue, 09 Nov 2004) $
 * $Author: ticso $
 * $Rev: 1795 $
 */

#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/uio.h>

#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include <netinet/in.h>
#include <netinet/tcp.h>

#define READ_COILS                              0x01
#define READ_DISCRETE_INPUTS                    0x02
#define READ_HOLDING_REGISTERS                  0x03
#define READ_INPUT_REGISTERS                    0x04
#define WRITE_SINGLE_COIL                       0x05
#define WRITE_SINGLE_REGISTER                   0x06
#define READ_EXCEPTION_STATUS                   0x07
#define DIAGNOSTICS                             0x08
#define GET_COMM_EVENT_COUNTER                  0x0b
#define GET_COMM_EVENT_LOG                      0x0c
#define WRITE_MULTIPLE_COILS                    0x0f
#define WRITE_MULTIPLE_REGISTERS                0x10
#define REPORT_SLAVE_ID                         0x11
#define READ_FILE_RECORD                        0x14
#define WRITE_FILE_RECORD                       0x15
#define MASK_WRITE_REGISTER                     0x16
#define READ_WRITE_MULTIPLE_REGISTERS           0x17
#define READ_FIFO_QUEUE                         0x18
#define ENCAPSULATED_INTERFACE_TRANSPORT        0x2b
# define READ_DEVICE_IDENTIFICATION             0x0d
#define VENDOR_SET_ADDRESS                      0x41
#define VENDOR_READ_MAGCARD                     0x42
#define VENDOR_PAGEDATA                         0x43
#define VENDOR_WRITE_PAGE                       0x44
#define VENDOR_RESET                            0x45
#define VENDOR_BOOTLOADER                       0x49

int main(int argc, char *argv[]);
void usage(void);

static int fd;
char *host;
char *port;

static uint8_t packet[256];
static uint8_t header[6];
static uint8_t packetlen;

uint16_t numbits;

ssize_t
readn(int fd, const void *vptr, size_t n) {
	char *ptr = (char*)vptr;
	fd_set fds;
	size_t nleft = n;
	while (nleft > 0) {
		FD_ZERO(&fds);
		FD_SET(fd, &fds);
		select(fd + 1, &fds, NULL, NULL, NULL);
		ssize_t nread;
		if ((nread = read(fd, ptr, nleft)) <= 0) {
			if (errno != EAGAIN && errno != EINTR)
				return(nread);
			nread = 0;
		}
		nleft -= nread;
		ptr += nread;
	}
	return (n);
}

ssize_t
writen(int fd, const void *vptr, size_t n) {
	char *ptr = (char*)vptr;
	fd_set fds;
	size_t nleft = n;
	while (nleft > 0) {
		FD_ZERO(&fds);
		FD_SET(fd, &fds);
		select(fd + 1, NULL, &fds, NULL, NULL);
		ssize_t nwritten;
		if ((nwritten = write(fd, ptr, nleft)) <= 0) {
			if (errno != EAGAIN && errno != EINTR)
				return(nwritten);
			nwritten = 0;
		}
		nleft -= nwritten;
		ptr += nwritten;
	}
	return (n);
}

int
sendpacket() {
	ssize_t len;

	if (fd == -1)
		return -1;
	header[0] = 0x00; // transaction identifier
	header[1] = 0x00;
	header[2] = 0x00; // protocol identifier
	header[3] = 0x00;
	header[4] = 0x00;
	header[5] = packetlen;
	len = writen(fd, header, sizeof(header));
	if (len != sizeof(header))
		return -1;
	len = writen(fd, packet, packetlen);
	if (len != packetlen)
		return -1;
	return 0;
}

void
getpacket() {
	readn(fd, header, sizeof(header));
	packetlen = header[5];
	readn(fd, packet, packetlen);
}

uint8_t getbits(int num, int pos) {
	uint8_t ret;
	int step;
	int i;

	if (num + pos > numbits)
		return 0;

	ret = 0;
	for (step = 0; step < num; step++) {
		ret = ret << 1;
		i = pos + step;
		if (packet[4 + (i / 8)] &
		    (0x80 >> (i & 7))) {
		} else {
			ret |= 0x01;
		}
	}
	return ret;
}

/*
 *      --Data Bits--   Parity
 *      b1  b2  b3  b4   b5    Character  Function<>
 *
 *      0   0   0   0    1        0 (0H)    Data
 *      1   0   0   0    0        1 (1H)      "
 *      0   1   0   0    0        2 (2H)      "
 *      1   1   0   0    1        3 (3H)      "
 *      0   0   1   0    0        4 (4H)      "
 *      1   0   1   0    1        5 (5H)      "
 *      0   1   1   0    1        6 (6H)      "
 *      1   1   1   0    0        7 (7H)      "
 *      0   0   0   1    0        8 (8H)      "
 *      1   0   0   1    1        9 (9H)      "
 *      0   1   0   1    1        : (AH)    Control
 *      1   1   0   1    0        ; (BH)    Start Sentinel
 *      0   0   1   1    1        < (CH)    Control
 *      1   0   1   1    0        = (DH)    Field Separator
 *      0   1   1   1    0        > (EH)    Control
 *      1   1   1   1    1        ? (FH)    End Sentinel<>
 *
 *         ***** 16 Character 5-bit Set *****
 *              10 Numeric Data Characters
 *              3 Framing/Field Characters
 *              3 Control Characters<>
 */
char decode(uint8_t val) {
	switch(val) {
	case 0x01:
		return '0';
	case 0x10:
		return '1';
	case 0x08:
		return '2';
	case 0x19:
		return '3';
	case 0x04:
		return '4';
	case 0x15:
		return '5';
	case 0x0d:
		return '6';
	case 0x1c:
		return '7';
	case 0x02:
		return '8';
	case 0x13:
		return '9';
	case 0x0b:
		return ':';
	case 0x1a:
		return ';';
	case 0x07:
		return '<';
	case 0x16:
		return '=';
	case 0x0e:
		return '>';
	case 0x1f:
		return '?';
	default:
		return '|';
	}
}

void
tcp_connect() {
	struct addrinfo hints, *res, *res0;
	int val;
	int error;
	const char *cause = NULL;

	fd = -1;
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	error = getaddrinfo(host, port, &hints, &res0);
	if (error) {
		printf("getaddrinfo failed: %s\n", gai_strerror(error));
		return;
	}
	cause = "no addresses";
	errno = EADDRNOTAVAIL;
	for (res = res0; res; res = res->ai_next) {
		fd = socket(res->ai_family, res->ai_socktype,
		res->ai_protocol);
		if (fd < 0) {
			cause = "socket";
			continue;
		}
		if (connect(fd, res->ai_addr, res->ai_addrlen) < 0) {
			cause = "connect";
			close(fd);
			fd = -1;
			continue;
		}
		break;  /* okay we got one */
	}
	if (fd < 0) {
		printf("%s failed\n", cause);
		printf("continuing anyway\n");
		return;
	}
	freeaddrinfo(res0);

	val = 1;
	setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &val, sizeof(val));
}

int
main(int argc, char *argv[]) {
	uint8_t address;
	int i;
	FILE *execpipe;
	char *readprog;
	uint8_t bitdata;
	uint8_t lrc;
	char num;

	if (argc != 5)
		usage();
	host = argv[1];
	port = argv[2];
	argc -= 3;
	argv += 3;

	tcp_connect();

	address = atol(argv[0]);
	readprog = argv[1];

	daemon(0, 0);

	for (;;) {
begin:
		packet[0] = address;
		packet[1] = VENDOR_READ_MAGCARD;
		packetlen = 2;
		if (sendpacket() < 0) {
			close(fd);
			sleep (1); // wait some time befor trying to reconnect
			tcp_connect();
			goto begin;
		}
		getpacket();
		if (packetlen == 0)
			goto begin;
		if (packet[1] & 0x80) {
			sleep(3); // XXX: device error - we should better inform the admin
			goto begin;
		}
		numbits = packet[2] << 8 | packet[3];
		lrc = 0x01;
		if (numbits != 0) {
			i = -1;
			do { // search for start char
				i++;
			} while (i < numbits && decode(getbits(5, i)) != ';');
			if (decode(bitdata = getbits(5, i)) != ';') {
				// no start character found
				execpipe = popen(readprog, "w");
				if (execpipe == NULL) {
					goto begin;
				}
				fprintf(execpipe, "+|?\n");
				pclose(execpipe);
				goto begin;
			}
			lrc ^= bitdata;
			i += 5;
			execpipe = popen(readprog, "w");
			if (execpipe == NULL) {
				goto begin;
			}
			fprintf(execpipe, "+");
			for (; i < numbits; i += 5) {
				bitdata = getbits(5, i);
				lrc ^= bitdata;
				num = decode(bitdata);
				if (num == '?') {
					if (i + 5 >= numbits) {
						fprintf(execpipe, "|no LRC|?");
					} else {
						// check LRC
						bitdata = getbits(5, i + 5);
						if (bitdata == lrc) {
							fprintf(execpipe, "?");
						} else {
							fprintf(execpipe, "|lrc 0x%x != 0x%x|?", lrc, bitdata);
						}
					}
					break;
				}
				fprintf(execpipe, "%c", num);
			}
			fprintf(execpipe, "\n");
			pclose(execpipe);
		}
		usleep(500000);
	}
	return 0;
}

void
usage(void) {

	printf("usage: mb_magcardd ip port slaveaddress script\n");
	exit(1);
}

